/*-----------------------------------------------------------------------*
 * filename - crtlvcl.cpp
 *
 * function(s)
 *
 *   __CRTL_VCL_Init
 *   __CRTL_VCL_Exit
 *
 *
 *   Both functions are called from the startup code.  Init happens
 *   very early and Exit happens very late.
 *
 *-----------------------------------------------------------------------*/

/*
 *      C/C++ Run Time Library - Version 15.0
 *
 *      Copyright (c) 1997, 2009 by Embarcadero Technologies, Inc.
 *      All Rights Reserved.
 *
 */

/* $Revision: 25283 $        */

#include <windows.h>
#include <process.h>
#include <_defs.h>
#include <_malloc.h>
#include <typeinfo.h>
#ifdef _BUILDRTLDLL
#undef _EXPDATA
#define _EXPDATA __declspec(package)  // for __pfnDli... import reference's
#endif
#include <delayimp.h>
#define USE_XX_VARS
#define DELPHI_EXCEPTIONS
#include <xx.h>

#undef __CPPexceptionList
#define __CPPexceptionList (__getExceptVarRec()->xvrCPPexceptionList)

///////////////////////////////////////////////////////////////////////////

struct TRaiseFrame
{
    void * NextRaise;
    void * ExceptAddr;
    void * ExceptObject;
    void * ExceptionRecord;
};

#define PASCALIMPLEMENTATION __declspec(pascalimplementation, package) // Implemented in Delphi - no code should be generated
#define DELPHICLASS          __declspec(delphiclass, package)          // Implemented in Delphi (C++ does not generate RTTI)

namespace System
{
    extern __declspec(package) HINSTANCE MainInstance;
    extern __declspec(package) bool      IsMultiThread;
    typedef void * (__fastcall * TSystemThreadFuncProc)(void *, void *);
    extern __declspec(package) TSystemThreadFuncProc SystemThreadFuncProc;
    typedef void (__fastcall * TSystemThreadEndProc)(int);
    extern __declspec(package) TSystemThreadEndProc SystemThreadEndProc;
    extern __declspec(package) void * __fastcall RaiseList(void);
    extern __declspec(package) void * __fastcall SetRaiseList(void * NewPtr);
    extern __declspec(package) void * ExceptObjProc;
    extern __declspec(package) void * RaiseExceptObjProc;
    extern __declspec(package) void * ExceptionAcquired;
    extern __declspec(package) void * UnloadDelayLoadedDLLPtr;
    extern __declspec(package) void * DelayLoadHelper;
    extern __declspec(package) void * pfnDliNotifyHook;
    extern __declspec(package) void * pfnDliFailureHook;
    extern void   __fastcall initialization (void);
    class DELPHICLASS TObject { };
};
using namespace System;

namespace Sysutils
{
    class DELPHICLASS Exception;
    class PASCALIMPLEMENTATION Exception : public System::TObject { };
    class DELPHICLASS EExternal;
    class PASCALIMPLEMENTATION EExternal : public Exception { };
    class DELPHICLASS EIntError;
    class PASCALIMPLEMENTATION EIntError : public EExternal { };
    class DELPHICLASS EDivByZero;
    class PASCALIMPLEMENTATION EDivByZero : public EIntError { };
    class DELPHICLASS ERangeError;
    class PASCALIMPLEMENTATION ERangeError : public EIntError { };
    class DELPHICLASS EIntOverflow;
    class PASCALIMPLEMENTATION EIntOverflow : public EIntError { };
    class DELPHICLASS EMathError;
    class PASCALIMPLEMENTATION EMathError : public EExternal { };
    class DELPHICLASS EInvalidOp;
    class PASCALIMPLEMENTATION EInvalidOp: public EMathError { };
    class DELPHICLASS EZeroDivide;
    class PASCALIMPLEMENTATION EZeroDivide: public EMathError { };
    class DELPHICLASS EOverflow;
    class PASCALIMPLEMENTATION EOverflow: public EMathError { };
    class DELPHICLASS EUnderflow;
    class PASCALIMPLEMENTATION EUnderflow: public EMathError { };
    class DELPHICLASS EAccessViolation;
    class PASCALIMPLEMENTATION EAccessViolation : public EExternal { };
    class DELPHICLASS EPrivilege;
    class PASCALIMPLEMENTATION EPrivilege : public EExternal { };
    class DELPHICLASS EStackOverflow;
    class PASCALIMPLEMENTATION EStackOverflow : public EExternal { };
    class DELPHICLASS EControlC;
    class PASCALIMPLEMENTATION EControlC : public EExternal { };
};
using namespace Sysutils;

///////////////////////////////////////////////////////////////////////////

typedef void * (_RTLENTRY * TGetExceptionObjectProc)(_EXCEPTION_RECORD *, tpidPtr *typeID);
typedef void (__fastcall * TRaiseExceptObjProc)(_EXCEPTION_RECORD *);
typedef void * (__fastcall * TExceptObjProc)(_EXCEPTION_RECORD *);

extern int _RTLENTRY _EXPFUNC setRaiseListFuncAddr(void *setAddr, void *clearAddr);
extern int _RTLENTRY _EXPFUNC setExceptionFuncAddr(TGetExceptionObjectProc getExceptionObject, TRaiseExceptObjProc *raiseExceptObjProc);
extern exceptVarPtr _RTLENTRY _EXPFUNC __getExceptVarRec(void);

/*
  This code is drawn primarily from the Delphi system unit for exception
  handling.  The behaviour is intended to parallel that of Delphi.
*/
static  unsigned char   mapException(EXCEPTION_RECORD *P)
{
    switch  (P->ExceptionCode)
    {
    case    STATUS_INTEGER_DIVIDE_BY_ZERO:  return 3;
    case    STATUS_ARRAY_BOUNDS_EXCEEDED:   return 4;
    case    STATUS_INTEGER_OVERFLOW:        return 5;

    case    STATUS_FLOAT_INEXACT_RESULT:
    case    STATUS_FLOAT_INVALID_OPERATION:
    case    STATUS_FLOAT_STACK_CHECK:       return 6;

    case    STATUS_FLOAT_DIVIDE_BY_ZERO:    return 7;
    case    STATUS_FLOAT_OVERFLOW:          return 8;

    case    STATUS_FLOAT_UNDERFLOW:
    case    STATUS_FLOAT_DENORMAL_OPERAND:  return 9;

    case    STATUS_ACCESS_VIOLATION:        return 11;
    case    STATUS_PRIVILEGED_INSTRUCTION:  return 12;
    case    STATUS_CONTROL_C_EXIT:          return 13;
    case    STATUS_STACK_OVERFLOW:          return 14;
    default:                                return 21;
    }
}

typedef struct  OSExceptionTypeID
{
    tpidPtr OSETid;
    unsigned char OSETexceptionNum;
} OSExceptionTypeID;

static  OSExceptionTypeID  OSExceptionTypeIDs[] =
{
    { 0,                                    0 },
    { 0,                                    1 },
    { 0,                                    2 },
    { typeid(EDivByZero       &).tpp,       3 },
    { typeid(ERangeError      &).tpp,       4 },
    { typeid(EIntOverflow     &).tpp,       5 },
    { typeid(EInvalidOp       &).tpp,       6 },
    { typeid(EZeroDivide      &).tpp,       7 },
    { typeid(EOverflow        &).tpp,       8 },
    { typeid(EUnderflow       &).tpp,       9 },
//  { typeid(EInvalidCast     *).tpp,       10 },
    { 0,                                    10 },
    { typeid(EAccessViolation &).tpp,       11 },
    { typeid(EPrivilege       &).tpp,       12 },
    { typeid(EControlC        &).tpp,       13 },
    { typeid(EStackOverflow   &).tpp,       14 },
};

static void *getExceptionObject(EXCEPTION_RECORD *P, tpidPtr *typeID)
{
    unsigned        ErrorCode;
    TObject *       excObj;
    TExceptObjProc	exceptObjProc;

    ErrorCode = mapException(P);

    if (ErrorCode < 3 || ErrorCode > sizeof(OSExceptionTypeIDs) / sizeof(OSExceptionTypeIDs[0]))
	return 0;

    *typeID = OSExceptionTypeIDs[ErrorCode].OSETid;

    exceptObjProc = (TExceptObjProc)System::ExceptObjProc;

    if (exceptObjProc)
	excObj = (TObject*)exceptObjProc(P);
    else
	excObj = NULL;

    return excObj;
}

static void __fastcall exceptionAcquired(TObject *e)
{
    if (e)
    {
	excDscPtr exc = __CPPexceptionList;
	while (exc)
	{
	    if ((exc->xdFlags & (XDF_ISDELPHIEXCEPTION|XDF_OSEXCEPTION)) && (*(TObject **)exc->xdValue == e))
	    {
		*(void **)exc->xdValue = NULL;
	    }
	    exc = exc->xdPrevious;
	}
    }
}

extern "C"
{

///////////////////////////////////////////////////////////////////////////

// These live in crtl_thd.c:

void _RTLENTRY _EXPFUNC __CRTL_VCL_Thread_Hook(void *pfunc, void *pend);
void _RTLENTRY _EXPFUNC __CRTL_VCL_Thread_Unhook(void *pfunc, void *pend);

///////////////////////////////////////////////////////////////////////////

void _RTLENTRY __CRTL_VCL_Sharemem (void);
extern unsigned char __isDLL;
extern unsigned char __isVCLPackage;

PEXCEPTION_RECORD _RTLENTRY __VCL_add_EH(PEXCEPTION_RECORD excPtr, void *);
void _RTLENTRY __VCL_clear_EH();
extern FARPROC WINAPI _EXPFUNC __delayLoadHelper(ImgDelayDescr *pidd, FARPROC *ppfnIATEntry);

///////////////////////////////////////////////////////////////////////////
// VCL Initializtion / Finalization Functions                            //
///////////////////////////////////////////////////////////////////////////

void _RTLENTRY __CRTL_VCL_Init (void)
{
    unsigned int hinst = _EDX;

    // We must set the VCL copy of hInstance (called MainInstance).
    // The startup code has placed this value into the EDX register for us.

    // Always set MainInstance in an EXE, and only set it in a DLL if it
    // hasn't been set.  Also, never set it in a PACKAGE.

    if (!__isVCLPackage)
        if ((!__isDLL) || (__isDLL && !System::MainInstance))
	    System::MainInstance = (HINSTANCE) hinst;


    // We must tell the Pascal RTL to enforce multithread issues in its
    // memory allocation routines, since we are in the multi-threaded RTL
    // on the C/C++ side.

    System::IsMultiThread = true;


#if 0 /* This won't be used in this update */
    // Callout to a function for Pascal thread hooking
    __CRTL_VCL_Thread_Hook(&System::SystemThreadFuncProc, &System::SystemThreadEndProc);
#endif

    // Now initialize the Pascal RTL

    System::initialization();
    setRaiseListFuncAddr(__VCL_add_EH, __VCL_clear_EH);
    setExceptionFuncAddr(getExceptionObject, (TRaiseExceptObjProc*)&System::RaiseExceptObjProc);
    System::ExceptionAcquired = &exceptionAcquired;
    System::UnloadDelayLoadedDLLPtr = &__FUnloadDelayLoadedDLL;
    System::DelayLoadHelper = &__delayLoadHelper;
    System::pfnDliNotifyHook = &__pfnDliNotifyHook;
    System::pfnDliFailureHook = &__pfnDliFailureHook;

#if defined(_BUILDRTLDLL)
    __CRTL_VCL_Sharemem();
#endif
}

///////////////////////////////////////////////////////////////////////////

void _RTLENTRY __CRTL_VCL_Exit (void)
{
// This cannot be performed now, since it prematurely unloads the memory
// manager DLL.  If the RTLDLL is being used, the RTL cannot free its memory
// without jumping into the non-existant memory manager DLL.  Therefore SYSTEM
// is initialized but never finalized.

//  #if defined(_BUILDRTLDLL)
//    System::Finalization();
//  #endif

#if 0 /* This won't be used in this update */
    // Callout to a function for Pascal thread unhooking
    __CRTL_VCL_Thread_Unhook(&System::SystemThreadFuncProc, &System::SystemThreadEndProc);
#endif
}

// Allocate memory for a Pascal exception and add it to RaiseList.
// This function is called from ___ExceptionHandler() while processing
// a Pascal exception code.

PEXCEPTION_RECORD _RTLENTRY __VCL_add_EH(PEXCEPTION_RECORD excPtr, void * obj)
{
    struct TRaiseFrame *myTRFrame;
    struct TRaiseFrame *curTRFrame = (TRaiseFrame *)System::RaiseList();
    if (excPtr)
    {
	myTRFrame = (TRaiseFrame *) __org_malloc(sizeof(TRaiseFrame));
	myTRFrame->ExceptAddr = excPtr->ExceptionAddress;
	myTRFrame->ExceptionRecord = excPtr;
	myTRFrame->ExceptObject = obj;
	myTRFrame->NextRaise = curTRFrame;
	System::SetRaiseList(myTRFrame);
    }
    return curTRFrame ? (PEXCEPTION_RECORD)curTRFrame->ExceptionRecord : 0;
}

// Remove the top exception object from RaiseList and deallocate it's memory.
// This function is called from _CatchCleanup() which verifies that this is
// a Pascal exception.

void _RTLENTRY __VCL_clear_EH()
{
    struct TRaiseFrame *myTRFrame = (TRaiseFrame *)System::RaiseList();

    if (myTRFrame)
    {
        System::SetRaiseList(myTRFrame->NextRaise);
        __org_free((void*)myTRFrame);
    }
}

//
// These functions start up and shut down VCL at the appropriate times.
//

void __stdcall __InitVCL(void);
void __stdcall __ExitVCL(void);

#pragma startup __InitVCL 28     // Initialization of early VCL routines.
                                 // Priority 28 is two eariler than what DCC32
                                 // uses for .PAS Units' Initialization
                                 // sections (which defaults to 30)

                                 // User's can now make a #29 priority startup
                                 // routine that will set IsLibrary = false
                                 // for statically linked DLLs that use VCL
                                 // Packages.

#pragma exit  __ExitVCL 31       // Finalization of VCL routines.
                                 // Priority 31 is one eariler than what DCC32
                                 // uses for .PAS Units' Finalization sections
                                 // (which defaults to 30)


///////////////////////////////////////////////////////////////////////////

} // extern "C"

///////////////////////////////////////////////////////////////////////////
